@@ -1,13 +1,15 @@
+import warnings
 import torch
 import torch.nn as nn
 from .odeint import SOLVERS, odeint
-from .misc import _check_inputs, _flat_to_shape, _rms_norm, _mixed_linf_rms_norm, _wrap_norm
+from .misc import _check_inputs, _flat_to_shape
+from .misc import _mixed_norm
 
 
 class OdeintAdjointMethod(torch.autograd.Function):
 
     @staticmethod
-    def forward(ctx, shapes, func, y0, t, rtol, atol, method, options, adjoint_rtol, adjoint_atol, adjoint_method,
+    def forward(ctx, shapes, func, y0, t, rtol, atol, method, options, event_fn, adjoint_rtol, adjoint_atol, adjoint_method,
                 adjoint_options, t_requires_grad, *adjoint_params):
 
         ctx.shapes = shapes
@@ -17,16 +19,23 @@ class OdeintAdjointMethod(torch.autograd.Function):
         ctx.adjoint_method = adjoint_method
         ctx.adjoint_options = adjoint_options
         ctx.t_requires_grad = t_requires_grad
+        ctx.event_mode = event_fn is not None
 
         with torch.no_grad():
-            y = odeint(func, y0, t, rtol=rtol, atol=atol, method=method, options=options)
+            ans = odeint(func, y0, t, rtol=rtol, atol=atol, method=method, options=options, event_fn=event_fn)
+
+            if event_fn is None:
+                y = ans
+            else:
+                event_t, y = ans
+                ctx.event_t = event_t
+
         ctx.save_for_backward(t, y, *adjoint_params)
-        return y
+        return ans
 
     @staticmethod
-    def backward(ctx, grad_y):
+    def backward(ctx, *grad_y):
         with torch.no_grad():
-            shapes = ctx.shapes
             func = ctx.func
             adjoint_rtol = ctx.adjoint_rtol
             adjoint_atol = ctx.adjoint_atol
@@ -37,35 +46,16 @@ class OdeintAdjointMethod(torch.autograd.Function):
             t, y, *adjoint_params = ctx.saved_tensors
             adjoint_params = tuple(adjoint_params)
 
-            ##################################
-            #     Set up adjoint_options     #
-            ##################################
-
-            if adjoint_options is None:
-                adjoint_options = {}
+            # Backprop as if integrating up to event time.
+            # Does NOT backpropagate through the event time.
+            event_mode = ctx.event_mode
+            if event_mode:
+                event_t = ctx.event_t
+                _t = t
+                t = torch.cat([t[0].reshape(-1), event_t.reshape(-1)])
+                grad_y = grad_y[1]
             else:
-                adjoint_options = adjoint_options.copy()
-
-            # We assume that any grid points are given to us ordered in the same direction as for the forward pass (for
-            # compatibility with setting adjoint_options = options), so we need to flip them around here.
-            try:
-                grid_points = adjoint_options['grid_points']
-            except KeyError:
-                pass
-            else:
-                adjoint_options['grid_points'] = grid_points.flip(0)
-
-            # Backward compatibility: by default use a mixed L-infinity/RMS norm over the input, where we treat t, each
-            # element of y, and each element of adj_y separately over the Linf, but consider all the parameters
-            # together.
-            if 'norm' not in adjoint_options:
-                if shapes is None:
-                    shapes = [y[-1].shape]  # [-1] because y has shape (len(t), *y0.shape)
-                # adj_t, y, adj_y, adj_params, corresponding to the order in aug_state below
-                adjoint_shapes = [torch.Size(())] + shapes + shapes + [torch.Size([sum(param.numel()
-                                                                                       for param in adjoint_params)])]
-                adjoint_options['norm'] = _mixed_linf_rms_norm(adjoint_shapes)
-            # ~Backward compatibility
+                grad_y = grad_y[0]
 
             ##################################
             #      Set up initial state      #
@@ -98,9 +88,9 @@ class OdeintAdjointMethod(torch.autograd.Function):
                     func_eval = func(t if t_requires_grad else t_, y)
 
                     # Workaround for PyTorch bug #39784
-                    _t = torch.as_strided(t, (), ())
-                    _y = torch.as_strided(y, (), ())
-                    _params = tuple(torch.as_strided(param, (), ()) for param in adjoint_params)
+                    _t = torch.as_strided(t, (), ())  # noqa
+                    _y = torch.as_strided(y, (), ())  # noqa
+                    _params = tuple(torch.as_strided(param, (), ()) for param in adjoint_params)  # noqa
 
                     vjp_t, vjp_y, *vjp_params = torch.autograd.grad(
                         func_eval, (t, y) + adjoint_params, -adj_y,
@@ -145,14 +135,18 @@ class OdeintAdjointMethod(torch.autograd.Function):
             if t_requires_grad:
                 time_vjps[0] = aug_state[0]
 
+            # Only compute gradient wrt initial time when in event handling mode.
+            if event_mode and t_requires_grad:
+                time_vjps = torch.cat([time_vjps[0].reshape(-1), torch.zeros_like(_t[1:])])
+
             adj_y = aug_state[2]
             adj_params = aug_state[3:]
 
-        return (None, None, adj_y, time_vjps, None, None, None, None, None, None, None, None, None, *adj_params)
+        return (None, None, adj_y, time_vjps, None, None, None, None, None, None, None, None, None, None, *adj_params)
 
 
-def odeint_adjoint(func, y0, t, rtol=1e-7, atol=1e-9, method=None, options=None, adjoint_rtol=None, adjoint_atol=None,
-                   adjoint_method=None, adjoint_options=None, adjoint_params=None):
+def odeint_adjoint(func, y0, t, *, rtol=1e-7, atol=1e-9, method=None, options=None, event_fn=None,
+                   adjoint_rtol=None, adjoint_atol=None, adjoint_method=None, adjoint_options=None, adjoint_params=None):
 
     # We need this in order to access the variables inside this module,
     # since we have no other way of getting variables along the execution path.
@@ -168,35 +162,64 @@ def odeint_adjoint(func, y0, t, rtol=1e-7, atol=1e-9, method=None, options=None,
         adjoint_atol = atol
     if adjoint_method is None:
         adjoint_method = method
+
+    if adjoint_method != method and options is not None and adjoint_options is None:
+        raise ValueError("If `adjoint_method != method` then we cannot infer `adjoint_options` from `options`. So as "
+                         "`options` has been passed then `adjoint_options` must be passed as well.")
+
     if adjoint_options is None:
         adjoint_options = {k: v for k, v in options.items() if k != "norm"} if options is not None else {}
+    else:
+        # Avoid in-place modifying a user-specified dict.
+        adjoint_options = adjoint_options.copy()
+
     if adjoint_params is None:
         adjoint_params = tuple(find_parameters(func))
     else:
         adjoint_params = tuple(adjoint_params)  # in case adjoint_params is a generator.
 
     # Filter params that don't require gradients.
+    oldlen_ = len(adjoint_params)
     adjoint_params = tuple(p for p in adjoint_params if p.requires_grad)
+    if len(adjoint_params) != oldlen_:
+        # Some params were excluded.
+        # Issue a warning if a user-specified norm is specified.
+        if 'norm' in adjoint_options and callable(adjoint_options['norm']):
+            warnings.warn("An adjoint parameter was passed without requiring gradient. For efficiency this will be "
+                          "excluded from the adjoint pass, and will not appear as a tensor in the adjoint norm.")
 
-    # Normalise to non-tupled input
-    shapes, func, y0, t, rtol, atol, method, options = _check_inputs(func, y0, t, rtol, atol, method, options, SOLVERS)
+    # Convert to flattened state.
+    shapes, func, y0, t, rtol, atol, method, options, event_fn, decreasing_time = _check_inputs(func, y0, t, rtol, atol, method, options, event_fn, SOLVERS)
 
-    if "norm" in options and "norm" not in adjoint_options:
-        adjoint_shapes = [torch.Size(()), y0.shape, y0.shape] + [torch.Size([sum(param.numel() for param in adjoint_params)])]
-        adjoint_options["norm"] = _wrap_norm([_rms_norm, options["norm"], options["norm"]], adjoint_shapes)
+    # Handle the adjoint norm function.
+    state_norm = options["norm"]
+    handle_adjoint_norm_(adjoint_options, shapes, state_norm)
 
-    solution = OdeintAdjointMethod.apply(shapes, func, y0, t, rtol, atol, method, options, adjoint_rtol, adjoint_atol,
-                                         adjoint_method, adjoint_options, t.requires_grad, *adjoint_params)
+    ans = OdeintAdjointMethod.apply(shapes, func, y0, t, rtol, atol, method, options, event_fn, adjoint_rtol, adjoint_atol,
+                                    adjoint_method, adjoint_options, t.requires_grad, *adjoint_params)
+
+    if event_fn is None:
+        solution = ans
+    else:
+        event_t, solution = ans
+        event_t = event_t.to(t)
+        if decreasing_time:
+            event_t = -event_t
 
     if shapes is not None:
         solution = _flat_to_shape(solution, (len(t),), shapes)
-    return solution
+
+    if event_fn is None:
+        return solution
+    else:
+        return event_t, solution
 
 
 def find_parameters(module):
 
     assert isinstance(module, nn.Module)
 
+    # If called within DataParallel, parameters won't appear in module.parameters().
     if getattr(module, '_is_replica', False):
 
         def find_tensor_attributes(module):
@@ -207,3 +230,51 @@ def find_parameters(module):
         return [param for _, param in gen]
     else:
         return list(module.parameters())
+
+
+def handle_adjoint_norm_(adjoint_options, shapes, state_norm):
+    """In-place modifies the adjoint options to choose or wrap the norm function."""
+
+    # This is the default adjoint norm on the backward pass: a mixed norm over the tuple of inputs.
+    def default_adjoint_norm(tensor_tuple):
+        t, y, adj_y, *adj_params = tensor_tuple
+        # (If the state is actually a flattened tuple then this will be unpacked again in state_norm.)
+        return max(t.abs(), state_norm(y), state_norm(adj_y), _mixed_norm(adj_params))
+
+    if "norm" not in adjoint_options:
+        # `adjoint_options` was not explicitly specified by the user. Use the default norm.
+        adjoint_options["norm"] = default_adjoint_norm
+    else:
+        # `adjoint_options` was explicitly specified by the user...
+        try:
+            adjoint_norm = adjoint_options['norm']
+        except KeyError:
+            # ...but they did not specify the norm argument. Back to plan A: use the default norm.
+            adjoint_options['norm'] = default_adjoint_norm
+        else:
+            # ...and they did specify the norm argument.
+            if adjoint_norm == 'seminorm':
+                # They told us they want to use seminorms. Slight modification to plan A: use the default norm,
+                # but ignore the parameter state
+                def adjoint_seminorm(tensor_tuple):
+                    t, y, adj_y, *adj_params = tensor_tuple
+                    # (If the state is actually a flattened tuple then this will be unpacked again in state_norm.)
+                    return max(t.abs(), state_norm(y), state_norm(adj_y))
+                adjoint_options['norm'] = adjoint_seminorm
+            else:
+                # And they're using their own custom norm.
+                if shapes is None:
+                    # The state on the forward pass was a tensor, not a tuple. We don't need to do anything, they're
+                    # already going to get given the full adjoint state as (t, y, adj_y, adj_params)
+                    pass  # this branch included for clarity
+                else:
+                    # This is the bit that is tuple/tensor abstraction-breaking, because the odeint machinery
+                    # doesn't know about the tupled nature of the forward state. We need to tell the user's adjoint
+                    # norm about that ourselves.
+
+                    def _adjoint_norm(tensor_tuple):
+                        t, y, adj_y, *adj_params = tensor_tuple
+                        y = _flat_to_shape(y, (), shapes)
+                        adj_y = _flat_to_shape(adj_y, (), shapes)
+                        return adjoint_norm((t, *y, *adj_y, *adj_params))
+                    adjoint_options['norm'] = _adjoint_norm
